/**********************************************************************
  GamessLog - parses GAMESS-US log files

  Copyright (C) 2009 Marcus D. Hanwell

  This file is part of the Avogadro molecular editor project.
  For more information, see <http://avogadro.sourceforge.net/>

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU Library General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
  02110-1301, USA.
 **********************************************************************/

#include "gamesslog.h"

#include "basisset.h"

#include <QFile>
#include <QTextStream>
#include <QStringList>
#include <QDebug>

using Eigen::Vector3d;
using std::vector;

namespace Avogadro
{
  using std::vector;

  GamessLog::GamessLog(QString filename, BasisSet* basis)
  {
    // Open the file for reading and process it
    QFile file(filename);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
      return;

    qDebug() << "File" << filename << "opened.";

    // Process the formatted checkpoint and extract all the information we need
    m_in = new QTextStream(&file);
    while (!m_in->atEnd()) {
      processLine();
    }

    // Now it should all be loaded load it into the basis set
//    load(basis);
  }

  GamessLog::~GamessLog()
  {
    delete m_in;
  }

  void GamessLog::processLine()
  {
    // First truncate the line, remove trailing white space and check
    QString line = m_in->readLine();
    QString key = line;

    // Big switch statement checking for various things we are interested in
    if (line.contains("ATOMIC                      COORDINATES (BOHR)")) {
      readAtoms();
      qDebug();
      for (int i = 0; i < m_atomNum.size(); ++i) {
        qDebug() << "Atom" << i << m_atomSymbol.at(i) << m_atomNum.at(i) << m_atomPos.at(i).x()
                 << m_atomPos.at(i).y() << m_atomPos.at(i).z();
      }
    }
    else if (line.contains("ATOMIC BASIS SET")) {
      readBasis();
      qDebug();
      for (int i = 0; i < m_shellToAtom.size(); ++i) {
        qDebug() << "Primitive" << i << m_shellNums.at(i) << m_primitiveTypes.at(i)
                 << m_a.at(i) << m_c.at(i) << m_csp.at(i);
      }
    }
    else if (line.contains("EIGENVECTORS")) {
      readMOs();
    }

  }

  void GamessLog::load(BasisSet* basis)
  {
    basis->setNormalized(true);
    basis->setAtomNum(m_atomNum);
    basis->setAtomPos(m_atomPos);
    basis->setPrimitiveTypes(m_primitiveTypes);
    basis->setShellIndices(m_shellNums);
    basis->setShellToAtom(m_shellToAtom);
    basis->setBasisExp(m_a);
    basis->setBasisCoeff(m_c);
    basis->setBasisCoeffsp(m_csp);
    basis->setOrbitalEnergy(m_orbitalEnergy);
    basis->setEigenVectors(m_eigenVectors);
  }

  void GamessLog::readAtoms()
  {
    QString line = m_in->readLine();
    while (!line.isEmpty()) {
      QStringList list = line.split(" ", QString::SkipEmptyParts);
      if (list.size() == 5) {
        m_atomSymbol.push_back(list.at(0));
        m_atomNum.push_back(list.at(1).toDouble());
        Vector3d vec(list.at(2).toDouble(),
                     list.at(3).toDouble(),
                     list.at(4).toDouble());
        m_atomPos.push_back(vec);
      }
      line = m_in->readLine();
    }
  }

  void GamessLog::readBasis()
  {
    // First skip to the lines that contain the basis set information
    QString line = m_in->readLine();
    while (!line.contains("SHELL TYPE"))
      line = m_in->readLine();

    // Now read in the basis set
    line = m_in->readLine();
    int currentAtom = -1;
    int currentShell = 0;
    while (!line.contains("TOTAL NUMBER OF BASIS SET SHELLS")) {
      QStringList list = line.split(" ", QString::SkipEmptyParts);
      if (list.size() == 1) {
        // Atom symbol
        if (list.at(0) != m_atomSymbol.at(++currentAtom))
          qDebug() << "Something weird - Atom symbols do not match.";
      }
      else if (list.size() >= 5) {
        // Line containing a primitive
        if (currentShell != list.at(0).toInt()) {
          // New shell
          m_shellToAtom.push_back(currentAtom);
          m_shellNums.push_back(m_a.size()-1);
        }
        m_primitiveTypes.push_back(primitiveType(list.at(1)));
        m_a.push_back(list.at(3).toDouble());
        m_c.push_back(list.at(4).toDouble());
        if (list.size() >= 6)
          m_csp.push_back(list.at(5).toDouble());
        else
          m_csp.push_back(0.0);
      }
      line = m_in->readLine();
    }

    // Now read in some of the extra data that follows
    while (!line.contains("$CONTRL OPTIONS")) {
      QStringList list = line.split("=", QString::SkipEmptyParts);
      line = m_in->readLine();
      if (list.size() != 2)
        continue;
      if (list.at(0).contains("TOTAL NUMBER OF BASIS SET SHELLS")) {
        if (m_shellNums.at(m_shellNums.size()-1) != list.at(1).toInt()) {
          qDebug() << "We are missing some shells"
                   << m_shellNums.at(m_shellNums.size()-1) << list.at(1).toInt();
        }
      }
      else if (list.at(0).contains("NUMBER OF CARTESIAN GAUSSIAN BASIS FUNCTIONS"))
        m_numBasisFunctions = list.at(1).toInt();
      else if (list.at(0).contains("NUMBER OF ELECTRONS"))
        m_numElectrons = list.at(1).toInt();
      else if (list.at(0).contains("CHARGE OF MOLECULE"))
        m_charge = list.at(1).toInt();
      else if (list.at(0).contains("TOTAL NUMBER OF ATOMS")) {
        if (m_atomPos.size() != list.at(1).toInt()) {
          qDebug() << "We are missing some atoms" << m_atomPos.size()
                   << list.at(1).toInt();
        }
      }
    }
    qDebug() << "Number of basis functions:" << m_numBasisFunctions;
  }

  void GamessLog::readMOs()
  {
    QString line = m_in->readLine();
    m_eigenVectors.resize(m_numBasisFunctions, m_numBasisFunctions);
    while (!line.contains("END OF RHF CALCULATION")) {
      QStringList list = line.split(" ", QString::SkipEmptyParts);
      bool haveEnergy;
      int currentMO;
      if (list.size() == 0) {
        currentMO = -1;
        haveEnergy = false;
      }
      else if (list.size() == 5) {
        // Header lines
        if (currentMO < 0)
          currentMO = list.at(0).toInt();
        else if (!haveEnergy) {
          foreach (QString str, list)
            m_orbitalEnergy.push_back(str.toDouble());
          haveEnergy = true;
        }
      }
      else if (list.size() == 9) {
        // Eigenvectors
        for (int i = 0; i < 5; ++i) {
          m_eigenVectors(currentMO + i, list.at(0).toInt()-1) = list.at(i+4).toDouble();
        }
      }
      line = m_in->readLine();
    }
  }

  int GamessLog::primitiveType(const QString &string)
  {
    if (string == "S")
      return BasisSet::S;
    else if (string == "L")
      return BasisSet::SP;
    else if (string == "P")
      return BasisSet::P;
    else if (string == "D")
      return BasisSet::D;
    else
      return -1;
  }

  vector<int> GamessLog::readArrayI(unsigned int n)
  {
    vector<int> tmp;
    while (tmp.size() < n)
    {
      QString line = m_in->readLine();
      QStringList list = line.split(" ", QString::SkipEmptyParts);
      for (int i = 0; i < list.size(); ++i)
        tmp.push_back(list.at(i).toInt());
    }
    return tmp;
  }

  vector<double> GamessLog::readArrayD(unsigned int n)
  {
    vector<double> tmp;
    while (tmp.size() < n)
    {
      QString line = m_in->readLine();
      QStringList list = line.split(" ", QString::SkipEmptyParts);
      for (int i = 0; i < list.size(); ++i)
        tmp.push_back(list.at(i).toDouble());
    }
    return tmp;
  }

  void GamessLog::outputAll()
  {
    qDebug() << "Shell mappings.";
    for (unsigned int i = 0; i < m_primitiveTypes.size(); ++i)
        qDebug() << i << ": type =" << m_primitiveTypes.at(i)
                      << ", number =" << m_shellNums.at(i)
                      << ", atom =" << m_shellToAtom.at(i);
    qDebug() << "MO coefficients.";
    for (unsigned int i = 0; i < m_MOcoeffs.size(); ++i)
      qDebug() << m_MOcoeffs.at(i);
  }

}
